/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/**
   \class AuFileIO
   \brief SoundFileIO for .au (.snd) soundfiles
   
   SoundFileIO reader/writer for .au (.snd) sound files.
   (Next, Sun audio format). 
*/



#include "AuFileIO.h"

/* Array containing descriptions of
the various formats for the samples
of the Next .snd/ Sun .au format */

char *format_array[8] = 
{
  "Unspecified format",
  "Mulaw 8-bit",
  "Linear 8-bit",
  "Linear 16-bit",   
  "Linear 24-bit",   
  "Linear 32-bit",   
  "Float",
  "Double"
};

/* sizes of the various formats in bytes */
short format_size_array[8] = {0,1,1,2,3,4,4,8};


AuFileIO::AuFileIO(string filename, int mode):SoundFileIO(filename, mode)
{
}


AuFileIO::~AuFileIO()
{
}

int 
AuFileIO::check_type()
{
  char magic[5];
  if (!sfp_)
    {
      cerr << "AuFileIO: File " << filename_ << " does not exist" << endl;
      return MRS_ERROR;
    }
  if (fread(magic, 1, 4, sfp_) == 0)
    {
      cerr << "AuFileIO: File " << filename_ << " is empty " << endl;
    }
  
  magic[4] = '\0';
  rewind(sfp_);
  if (!strcmp(magic, ".snd"))
    {
      return MRS_OK;
    }
  else return MRS_ERROR;

}



void 
AuFileIO::init()
{
  if (mode_ == MRS_SF_READ)
    {
      // Read Header 
      fread(&hdr_, sizeof(snd_header), 1, sfp_);
  
      // Find out length of comments 
      hdr_.hdrLength = ntohl(hdr_.hdrLength);
      hdr_.comment[hdr_.hdrLength-24] = '\0';
      id_ = 15;
      hdr_.srate = ntohl(hdr_.srate);
      hdr_.channels = ntohl(hdr_.channels);
      hdr_.mode = ntohl(hdr_.mode);
      hdr_.fileLength = ntohl(hdr_.fileLength);
      srate_ = hdr_.srate;
      id_ = 15;
      winSize_ = DEFAULT_WIN_SIZE;
      hopSize_ = DEFAULT_HOP_SIZE;
      skipSize_ = DEFAULT_SKIP_SIZE;
      zeroSize_ = DEFAULT_ZERO_SIZE;
      channels_ = hdr_.channels;
      sizeSamples_ = (hdr_.fileLength) / format_size_array[hdr_.mode] / hdr_.channels;
      sizeSeconds_ = sizeSamples_ / srate_;
      comments_    = hdr_.comment;
      type_ = "Au (.snd) file IO sound object";
      fseek(sfp_, hdr_.hdrLength, 0);
      sfp_begin_ = ftell(sfp_);
    }
  else if (mode_ == MRS_SF_WRITE)
    {
      written_ = 0;
      char *comment = "MARSYAS 2001, George Tzanetakis.\n";
      int comment_size = strlen(comment);
      sfp_ = fopen(filename_.c_str(), "wb");
      hdr_.pref[0] = '.';
      hdr_.pref[1] = 's';
      hdr_.pref[2] = 'n';
      hdr_.pref[3] = 'd';
      hdr_.hdrLength = htonl(24 + comment_size);
      hdr_.fileLength = htonl(0);
      hdr_.mode = htonl(SND_FORMAT_LINEAR_16);                           
      hdr_.srate = htonl(srate_);
      hdr_.channels = htonl(channels_);
      fwrite(&hdr_, 24, 1, sfp_);
      // Write comment part of header 
      fwrite(comment, comment_size, 1, sfp_);      
      sfp_begin_ = ftell(sfp_);
    }
}

float*
AuFileIO::getall()
{
  float *all = new float[sizeSamples_];
  fvec win;
  win.create(winSize_);
  int i, j, windows = sizeSamples_/winSize_;
    switch (hdr_.mode)
    {
    case SND_FORMAT_UNSPECIFIED:
      {
        cerr << "Unspecified format \n";
        break;
      }
    case SND_FORMAT_MULAW_8:
      {
        cerr << "MU_LAW for now not supported \n";
        break;
      }
    case SND_FORMAT_LINEAR_8:
      {
	for(i=0; i<windows; i++)
	{
          read_linear8(win, i);
          for(j = 0; j < winSize_; j++) all[i*winSize_+j] = win(j);
	}
	break;
      }
    case SND_FORMAT_LINEAR_16:
      {
        for(i=0; i<windows; i++)
        { 
          read_linear16Swap(win, i);
          for(j = 0; j < winSize_; j++) all[i*winSize_+j] = win(j);
        }
        break;
      }
    case SND_FORMAT_FLOAT:
      {
        for(i=0; i<windows; i++)
        { 
          read_float(win, i);
          for(j = 0; j < winSize_; j++) all[i*winSize_+j] = win(j);
        }
        break;
      }
    default:
      {
        cerr << "File mode " << format_array[hdr_.mode] << "(" << hdr_.mode << ") is not supported for now\n";
      }
    }
}


void
AuFileIO::get(fvec &win, unsigned long index)
{
  if (win.size() != winSize_) 
    {
      cerr << "AuFileIO::get_win incompatible window sizes\n" << endl;
      cerr << "win.size() = " << win.size() << endl;
      cerr << "winSize_ = " << winSize_ << endl;
      exit(1);
    }
  
  switch (hdr_.mode)
    {
    case SND_FORMAT_UNSPECIFIED:
      {
	cerr << "Unspecified format \n";
	break;
      }
    case SND_FORMAT_MULAW_8:
      {
	cerr << "MU_LAW for now not supported \n";
	break;
      }
    case SND_FORMAT_LINEAR_8:
      {
	read_linear8(win, index);
	break;
      }
    case SND_FORMAT_LINEAR_16:
      {
	read_linear16Swap(win, index);
	break;
      }
    case SND_FORMAT_FLOAT:
      {
	read_float(win, index);
	break;
      }
    default:
      {
	cerr << "File mode " << format_array[hdr_.mode] << "(" << hdr_.mode << ") is not supported for now\n";
      }
    }
}


void 
AuFileIO::put(fvec &win, unsigned long index)
{
  if (win.size() != winSize_) 
    {
      cerr << "AuFileIO::put incompatible window sizes\n" << endl;
      exit(1);
    }

  long fileSize;
  pos_ = ftell(sfp_);  
  fseek(sfp_, 8, SEEK_SET);
  written_ += win.size();
  fileSize = htonl(written_ * 2 * channels_);
  fwrite(&fileSize, 4, 1, sfp_);
  fseek(sfp_, pos_, SEEK_SET);
  write_linear16(win, index);
}

  
  


