/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/** 
    \class SFM
    \brief SFM energy of window

   Calculate the SFM energy of the input fvec window. 
SFM is defined as the square root of the sum of 
the squarred values of the signal. 
*/



#include "SFM.h"
#include "math.h"

SFM::SFM()
{
  inSize_ = DEFAULT_WIN_SIZE;
  outSize_ = 1;
  init();
}

SFM::SFM(unsigned int inSize)
{
  inSize_ = inSize;
  outSize_ = 1;
  init();
}

SFM::~SFM()
{
}

void
SFM::init()
{
  magfft_ = new MagFFT(inSize_);
  bands_ = 25;
  fft_.create(inSize_/2);
  bandbounds_.create(bands_+1);

  bandbounds_(0) = 0;
  bandbounds_(1) = 100;
  bandbounds_(2) = 200;
  bandbounds_(3) = 300;
  bandbounds_(4) = 400;
  bandbounds_(5) = 510;
  bandbounds_(6) = 630;
  bandbounds_(7) = 770;
  bandbounds_(8) = 920;
  bandbounds_(9) = 1080;
  bandbounds_(10) = 1270;
  bandbounds_(11) = 1480;
  bandbounds_(12) = 1720;
  bandbounds_(13) = 2000;
  bandbounds_(14) = 2320;
  bandbounds_(15) = 2700;
  bandbounds_(16) = 3150;
  bandbounds_(17) = 3700;
  bandbounds_(18) = 4400;
  bandbounds_(19) = 5300;
  bandbounds_(20) = 6400;
  bandbounds_(21) = 7700;
  bandbounds_(22) = 9500;
  bandbounds_(23) = 12000;
  bandbounds_(24) = 15500;
  bandbounds_(25) = 22050;
 
  bandbounds_ *= inSize_/44100.0; 
}

void 
SFM::process(fvec& in, fvec& out) 
{
  if ((in.size() != inSize_) || (out.size() != outSize_))
  {
      cerr << "Warning: SFM::process: inSize_ and input window size do not agree" << endl;
      return;
  }
  
  magfft_->process(in, fft_);
  
  // calculate energies for each band
  int i, j;
  double energy[bands_];
  for(i = 0; i < bands_; i++)
    for(j = (int)floor(bandbounds_(i)), energy[i] = 0.0; j < (int)floor(bandbounds_(i+1)); j++)
	  energy[i] += fft_(j);
	
  // calcualte geometric and arithmetic means
  double gmean = 1.0, amean = 0;
  for(i = 0; i < bands_; i++)
  {
    gmean *= energy[i];
	amean += energy[i];
  }
  gmean = pow(gmean, 1.0/(float)bands_);
  amean = amean / bands_;
  
  // calculate sfm

  out(0) = amean;
  if(out(0) != 0) out(0) = gmean / out(0);
  if(out(0) != 0) out(0) = -log10(out(0)) / 6;
  if(out(0) > 1) out(0) = 1;
}


	

	
