/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/** 
    \class Series
    \brief Series of System objects
    
    Combines a series of System objects to a single System 
corresponding to executing the System objects one after the other 
in sequence. 
*/




#include "Series.h"

Series::Series()
{
  numFeats_ = 0;
  inSize_ =  DEFAULT_WIN_SIZE;
  outSize_ = DEFAULT_WIN_SIZE;
  vectors_ = NULL;
  featArray_.reserve(10);
}

Series::Series(vector<System*> featArray)
{
  numFeats_ = featArray.size();
  vectors_ = new fvec[numFeats_];
  featArray_ = featArray;
  
  for (unsigned int i=0; i<numFeats_-1; i++)
    {
      if (featArray_[i]->outSize() != featArray_[i+1]->inSize())
	{
	  cerr << "Cannot create feature series. Check inSize and outSize for features" << endl;
	  numFeats_ = 0;
	}
      vectors_[i].allocate(featArray_[i]->outSize());
    }
  inSize_ = featArray_[0]->inSize();
  outSize_ = featArray_[numFeats_-1]->outSize();
}

Series::~Series()
{
  delete [] vectors_;
}


void
Series::process(fvec& in, fvec& out)
{
  if ((in.size() != inSize_) || (out.size() != outSize_))
    {
      cerr << "Warning: Series::process: inSize_ and input window size or outSize_ and output window size do not agree" << endl;
      cerr << "inSize_ = " << inSize_ << endl;
      cerr << "input window size = " << in.size() << endl;
      cerr << "outSize_ = " << outSize_ << endl;
      cerr << "output window size = " << out.size() << endl;
      return;
    }
  
  for (unsigned int i=0; i<numFeats_; i++)
    {
      if (i==0)
	featArray_[i]->process(in, vectors_[i]);
      else if (i == numFeats_-1)
	featArray_[i]->process(vectors_[i-1], out);
      else
	featArray_[i]->process(vectors_[i-1], vectors_[i]);
    }
}
