/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


/**
   \class Spectral1
   \brief FFT based Spectral1 features 


   Calculate FFT mangitude based features. 
The Centroid, Rolloff and Flux are calculated 
on the magnitude spectrum of the input fvec 
and Zcrs are calculated in time domain. 
This System feature are typically used in classification. 
*/


#include "Spectral1.h"

Spectral1::Spectral1()
{
  inSize_ = DEFAULT_WIN_SIZE;
  zeroSize_ = 0;
  outSize_ = 4;
  init();
}

Spectral1::Spectral1(unsigned int inSize, unsigned int zeroSize)
{
  inSize_ = inSize;
  zeroSize_ = zeroSize;
  outSize_ = 4;
  init();
}


void 
Spectral1::init()
{
  temp0_.create(3);  
  temp1_.create(1);
  temp2_.create(1);
  temp3_.create(4);
  temp4_.create(8);
  temp5_.create(2);
  temp6_.create(1);
  
  
  hamming_ = new Hamming(inSize_, zeroSize_);
  magfft_  = new MagFFT(inSize_);
  shapeDesc_  = new ShapeDesc(magfft_->outSize());
  rms_ = new RMS(inSize_);
  

  vector<System *> featArray;
  featArray.push_back(hamming_);
  featArray.push_back(magfft_);
  featArray.push_back(shapeDesc_);
  series_ = new Series(featArray);

  
  featSize_ = outSize_;
  featNames_.push_back("Centroid");
  featNames_.push_back("Rolloff");
  featNames_.push_back("Flux");
  featNames_.push_back("RMD");
}



Spectral1::~Spectral1()
{
  delete hamming_;
  delete magfft_;
  delete shapeDesc_;
  delete rms_;
  delete series_;
}


void 
Spectral1::process(fvec& in, fvec& out) 
{
  unsigned int i;
  assert((in.size() == inSize_) && (out.size() == outSize_));      
  
  series_->process(in, temp0_);
  rms_->process(in, temp2_);
	
	out(0) = temp0_(1);
  out(1) = temp2_(0); 
	out(2) = temp0_(0);
	out(3) = temp0_(2);
}


	

	
