/*
** Copyright (C) 2000 George Tzanetakis <gtzan@cs.princeton.edu>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/**
    \class WavFileIO
    \brief SoundFileIO for .au (.snd) soundfiles
 
   SoundFileIO reader/writer for .wav sound files.

*/



#include "WavFileIO.h"





WavFileIO::WavFileIO(string filename, int mode):SoundFileIO(filename, mode)
{
}


WavFileIO::~WavFileIO()
{
}

int 
WavFileIO::check_type()
{
   char magic[5];
   if (!sfp_)
    {
      cerr << "WavFileIO: File " << filename_ << " does not exist" << endl;
      return MRS_ERROR;
    }
   fseek(sfp_, 8, SEEK_SET); // Locate wave id
   if (fread(magic, 4, 1, sfp_) == 0)
    {
      cerr << "WavFileIO: File " << filename_ << " is empty " << endl;
    }
  
  magic[4] = '\0';
  rewind(sfp_);
  if (!strcmp(magic, "WAVE"))
    {
      cerr << "File " << filename_ << " is WAVE (.wav) format \n" << endl;
      return MRS_OK;
    }
  else return MRS_ERROR;
}


void 
WavFileIO::init()
{
  cerr << "WavFileIO::init() " << endl;
  char id[5];
  fseek(sfp_, 12, SEEK_SET); // Locate wave id  
  if (mode_ == MRS_SF_READ)
    {
      
      int chunkSize;
      fread(id, 4, 1, sfp_);
      id[4] = '\0';
      cerr << "id = " << id << endl;
      while (strcmp(id, "fmt ")) 
	{
	  fread(&chunkSize, 4, 1, sfp_);
	  //chunkSize = ntohl(chunkSize);
	  fseek(sfp_, chunkSize, SEEK_CUR);
	  fread(id, 4, 1, sfp_);
	  cerr << "ChunkSize = " << chunkSize << endl;
	}
      cerr << "fmt chunk discovered " << endl;
      fread(&chunkSize, 4, 1, sfp_);

      unsigned short format_tag;
      fread(&format_tag, 2, 1, sfp_);
      //format_tag = ntohs(format_tag);
      cerr << "Format tag = " << format_tag << endl;
      if (format_tag != 1) 
	{
	  fclose(sfp_);
	  cerr << "Non pcm(compressed) wave files are not supported" << endl;
	  exit(1);
	}
      
      // Get number of channels
      unsigned short channels;      
      fread(&channels, 2,1, sfp_);
      channels_ = channels;
      cerr << "Channels = " << channels_ << endl;
      

      unsigned short srate;
      fread(&srate, 2,1,sfp_);
      srate_ = srate;
      cerr << "Srate = " << srate_ << endl;
      
      fseek(sfp_,8,SEEK_CUR);
      fread(&bits_, 2, 1, sfp_);
      //bits_ = ntohs(bits_);
      
      
      
      if ((bits_ != 16)&&(bits_ != 8)) 
	{
	  cerr << "WavFileIO::Only linear 8-bit and 16-bit samples are supported " << endl;
	  cerr << "Bits = " << bits_ << endl;
	  exit(1);
	}
      fseek(sfp_, chunkSize - 16, SEEK_CUR);
      

      fread(id, 4, 1, sfp_);
      id[4] = '\0';
      cerr << "Bits = " << bits_ << endl;
      cerr << "id = " << id << endl;
      
      while (strcmp(id, "data"))
	{
	  fread(&chunkSize, 4, 1, sfp_);
	  chunkSize = ntohl(chunkSize);
	  cout << "ChunkSize = " << chunkSize << endl;
	  fseek(sfp_,chunkSize,SEEK_CUR);
	  fread(&id,4,1,sfp_);	  
	  cerr << "id = " << id << endl;
	}

      int bytes;
      fread(&bytes, 4, 1, sfp_);
      //bytes = ntohl(bytes);
      

      sizeSamples_ = bytes / (bits_ / 8)/ channels_;
      sizeSeconds_ = sizeSamples_ / srate_;
      id_ = 16;
      winSize_ = DEFAULT_WIN_SIZE;
      hopSize_ = DEFAULT_HOP_SIZE;
      skipSize_ = DEFAULT_SKIP_SIZE;
      zeroSize_ = DEFAULT_ZERO_SIZE;
      comments_    = "George Tzanetakis, MARSYAS 1998";
      type_ = "Wave (.wav) file IO sound object";      
      sfp_begin_ = ftell(sfp_);

    }
  else if (mode_ == MRS_SF_WRITE)
    {
      /* written_ = 0;
      char *comment = "MARSYAS 2001, George Tzanetakis.\n";

      int comment_size = strlen(comment);
      sfp_ = fopen(filename_.c_str(), "wb");
      hdr_.pref[0] = '.';
      hdr_.pref[1] = 's';
      hdr_.pref[2] = 'n';
      hdr_.pref[3] = 'd';
      hdr_.hdrLength = htonl(24 + comment_size);
      hdr_.fileLength = htonl(0);
      hdr_.mode = htonl(SND_FORMAT_LINEAR_16);                           
      hdr_.srate = htonl(srate_);
      hdr_.channels = htonl(channels_);
      fwrite(&hdr_, 24, 1, sfp_);
      fwrite(comment, comment_size, 1, sfp_);      
      sfp_begin_ = ftell(sfp_);
      */
    }
  //cerr << "WavFileIO::init() end" << endl;
}


void
WavFileIO::get(fvec &win, unsigned long index)
{
  if (win.size() != winSize_) 
    {
      cerr << "WavFileIO::get_win incompatible window sizes\n" << endl;
      exit(1);
    }
  switch(bits_) 
    {
    case 16: 
      {
	read_linear16(win, index);
	break;
      }
    case 8:
      {
	read_linear8(win, index);
	break;
      }
    }
}


void 
WavFileIO::put(fvec &win, unsigned long index)
{
  /*
  if (win.size() != winSize_) 
    {
      cerr << "WavFileIO::put incompatible window sizes\n" << endl;
      exit(1);
    }

  long fileSize;
  pos_ = ftell(sfp_);  
  fseek(sfp_, 8, SEEK_SET);
  cout << "written = " << written_<< endl;
  written_ += win.size();
  fileSize = htonl(written_ * 2 * channels_);
  fwrite(&fileSize, 4, 1, sfp_);
  fseek(sfp_, pos_, SEEK_SET);
  write_linear16(win, index);
  */
}

  
  


